@@ -1,6 +1,7 @@
 # Copyright (C) 2021 Sebastian Pipping <sebastian@pipping.org>
 # Licensed under GNU Affero GPL version 3 or later
 
+import base64
 import datetime
 import os
 import re
@@ -12,6 +13,8 @@ import tempfile
 from argparse import ArgumentParser
 from contextlib import suppress
 
+import yaml
+
 from ..atoms import ATOM_LIKE_DISPLAY, extract_category_package_from
 from ..reporter import announce_and_call, announce_and_check_output, exception_reporting
 from ..version import VERSION_STR
@@ -41,8 +44,8 @@ def enrich_config(config):
 
     config.host_logdir = os.path.realpath(config.host_logdir)
     config.host_etc_portage = os.path.realpath(config.host_etc_portage)
-    if config.host_tweaks_dir is not None:
-        config.host_tweaks_dir = os.path.realpath(config.host_tweaks_dir)
+    if config.host_flavors_dir is not None:
+        config.host_flavors_dir = os.path.realpath(config.host_flavors_dir)
 
     if config.cflags is None:
         config.cflags = announce_and_check_output(['portageq', 'envvar', 'CFLAGS']).rstrip()
@@ -134,10 +137,11 @@ def parse_command_line(argv):
                         default='/etc/portage',
                         help='enforce specific location for /etc/portage (default: "%(default)s")')
 
-    parser.add_argument('--tweaks',
-                        dest='host_tweaks_dir',
+    parser.add_argument('--flavors',
+                        dest='host_flavors_dir',
                         metavar='DIR',
-                        help='location of directory containing tweak files (e.g. package.use)')
+                        help=('location of directory containing '
+                              'sparse <category>/<package>/flavors.yml file hierarchy'))
 
     parser.add_argument('--shy-rebuild',
                         dest='enforce_rebuild',
@@ -204,7 +208,6 @@ def build(config):
     emerge = ['env'] + emerge_env + ['emerge'] + emerge_args
     emerge_quoted_flat = ' '.join(emerge)
     rebuild_or_not = f'--usepkg={"n" if config.enforce_rebuild else "y"}'
-    install_or_not = '' if config.enforce_installation else '--buildpkgonly'
 
     container_profile_dir = os.path.join(container_portdir, 'profiles', config.gentoo_profile)
     container_command_shared_prefix = [
@@ -217,18 +220,9 @@ def build(config):
         f'ln -f -s {shlex.quote(container_profile_dir)} /etc/make.profile',
     ]
 
-    container_command_flat_check = ' && '.join(container_command_shared_prefix + [
-        f'{emerge_quoted_flat} --usepkg=y --onlydeps --pretend --verbose-conflicts {shlex.quote(config.atom)}',  # noqa: E501
-    ])
-
-    container_command_flat_build = ' && '.join(container_command_shared_prefix + [
-        f'{emerge_quoted_flat} --usepkg=y --onlydeps {shlex.quote(config.atom)}',
-        f'{emerge_quoted_flat} {rebuild_or_not} {install_or_not} {shlex.quote(config.atom)}',
-    ])
-
     # Create pretend log dir
     category, package = extract_category_package_from(config.atom)
-    host_pretend_logdir = os.path.join(config.host_logdir, 'pretend')
+    host_pretend_logdir = os.path.join(config.host_logdir, 'binary-gentoo')
     host_pretend_logdir_category = os.path.join(host_pretend_logdir, category)
     host_pretend_logdir_category_package = os.path.join(host_pretend_logdir_category, package)
     os.makedirs(host_pretend_logdir_category_package, mode=0o700, exist_ok=True)
@@ -239,63 +233,111 @@ def build(config):
         ]
         announce_and_call(rsync_argv)
 
-        if config.host_tweaks_dir is not None:
-            global_package_use_file = os.path.join(config.host_tweaks_dir, 'package.use')
-            single_package_use_file = os.path.join(config.host_tweaks_dir, category, package,
-                                                   'package.use')
-            for package_use_file, target_basename in (
-                (global_package_use_file, 'ZZ-global'),  # to be second last in alphabetic order
-                (single_package_use_file,
-                 f'ZZZ-{category}-{package}'),  # to be last in alphabetic order
-            ):
-                if os.path.exists(package_use_file):
-                    target_package_use_file = os.path.join(eventual_etc_portage, 'package.use',
-                                                           target_basename)
-                    shutil.copyfile(package_use_file, target_package_use_file)
+        os.makedirs(os.path.join(eventual_etc_portage, 'package.use'), exist_ok=True)
 
-        filename_timestamp = str(datetime.datetime.now()).replace(' ', '-').replace(':',
-                                                                                    '-').replace(
-                                                                                        '.', '-')
-        pretend_fail_log_filename = os.path.join(host_pretend_logdir_category_package,
-                                                 filename_timestamp + '-fail.log')
-        pretend_log_writer_process = subprocess.Popen(['tee', pretend_fail_log_filename],
-                                                      stdin=subprocess.PIPE)
-        try:
-            for container_command_flat, stdout, remove_log_file in (
-                (container_command_flat_check, pretend_log_writer_process.stdin, True),
-                (container_command_flat_build, None, False),
-            ):
-                docker_run_args = [
-                    '--rm',
-                    '-v',
-                    f'{eventual_etc_portage}:/etc/portage:rw',
-                    '-v',
-                    f'{config.host_logdir}:{container_logdir}:rw',
-                    '-v',
-                    f'{config.host_portdir}:{container_portdir}:ro',
-                    '-v',
-                    f'{config.host_pkgdir}:/var/cache/binpkgs:rw',
-                    '-v',
-                    f'{config.host_distdir}:/var/cache/distfiles:rw',
-                    config.docker_image,
-                    'sh',
-                    '-c',
-                    container_command_flat,
+        # Write package.use level "global", 1 of 4
+        flavors_yml_doc = {}
+        if config.host_flavors_dir is not None:
+            global_package_use_file_source = os.path.join(config.host_flavors_dir, 'package.use')
+            if os.path.exists(global_package_use_file_source):
+                global_package_use_file_target = os.path.join(
+                    eventual_etc_portage, 'package.use',
+                    'ZZ-global')  # to be third last in alphabetic order
+                shutil.copyfile(global_package_use_file_source, global_package_use_file_target)
+
+            flavors_yml_file = os.path.join(config.host_flavors_dir, category, package,
+                                            'flavors.yml')
+            with suppress(OSError), open(flavors_yml_file) as f:
+                # TODO validate flavors.yml file content
+                flavors_yml_doc = yaml.safe_load(f)
+
+        # Write package.use level "common to all flavors", 2 of 4
+        common_package_use_file_content = flavors_yml_doc.get('package.use')
+        if common_package_use_file_content:
+            common_package_use_file_target = os.path.join(
+                eventual_etc_portage, 'package.use',
+                'ZZZ-common')  # to be second last in alphabetic order
+            with open(common_package_use_file_target, 'w') as f:
+                print(common_package_use_file_content, file=f)
+
+        # Build flavors
+        for flavor in flavors_yml_doc.get('flavors', [{}]):
+
+            # Write package.use level "single flavor", 3 of 4
+            # NOTE: We're overwriting every time to not have package.use of the previous flavor
+            #       affect the current build
+            single_package_use_file_target = os.path.join(
+                eventual_etc_portage, 'package.use',
+                'ZZZZ-flavor')  # to be last in alphabetic order
+            with open(single_package_use_file_target, 'w') as f:
+                print(flavor.get('package.use', ''), file=f)
+
+            filename_timestamp = str(datetime.datetime.now()).replace(' ', '-').replace(
+                ':', '-').replace('.', '-')
+            pretend_fail_log_filename = os.path.join(host_pretend_logdir_category_package,
+                                                     filename_timestamp + '-fail.log')
+            pretend_log_writer_process = subprocess.Popen(['tee', pretend_fail_log_filename],
+                                                          stdin=subprocess.PIPE)
+
+            # Assemble build command list
+            steps = flavor.get('steps', [{}])
+            step_commands = container_command_shared_prefix
+            prior_step_package_use_file_exists = False
+            for step_index, step in enumerate(steps):
+                is_last_step = step_index == len(steps) - 1
+
+                # Re-write step-specific package.use file, 4 of 4
+                package_use_content = base64.encodebytes(
+                    step.get('package.use', '').encode('ascii')).decode("ascii").replace('\n', '')
+                flavor_package_use_file = '/etc/portage/package.use/ZZZZZ-step'  # to be very last
+                if package_use_content:
+                    step_commands.append(
+                        f'base64 -d <<<{package_use_content!r} | tee {flavor_package_use_file}')
+                    prior_step_package_use_file_exists = True
+                elif prior_step_package_use_file_exists:
+                    step_commands.append(f'rm -f {flavor_package_use_file}')  # from previous step
+                    prior_step_package_use_file_exists = False
+
+                install_or_not = ('' if (config.enforce_installation or not is_last_step) else
+                                  '--buildpkgonly')
+                step_commands += [
+                    f'{emerge_quoted_flat} --usepkg=y --onlydeps --verbose-conflicts {shlex.quote(config.atom)}',  # noqa: E501
+                    f'{emerge_quoted_flat} {rebuild_or_not} {install_or_not} {shlex.quote(config.atom)}',  # noqa: E501
                 ]
 
-                if config.interactive:
-                    docker_run_args = ['-it'] + docker_run_args
+            container_command_flat = ' && '.join(step_commands)
 
-                announce_and_call(['docker', 'run'] + docker_run_args, stdout=stdout)
+            docker_run_args = [
+                '--rm',
+                '-v',
+                f'{eventual_etc_portage}:/etc/portage:rw',
+                '-v',
+                f'{config.host_logdir}:{container_logdir}:rw',
+                '-v',
+                f'{config.host_portdir}:{container_portdir}:ro',
+                '-v',
+                f'{config.host_pkgdir}:/var/cache/binpkgs:rw',
+                '-v',
+                f'{config.host_distdir}:/var/cache/distfiles:rw',
+                config.docker_image,
+                'sh',
+                '-c',
+                container_command_flat,
+            ]
 
-                if remove_log_file:
-                    os.remove(pretend_fail_log_filename)
-                    with suppress(OSError):
-                        os.rmdir(host_pretend_logdir_category_package)
-                        os.rmdir(host_pretend_logdir_category)
-        finally:
-            pretend_log_writer_process.stdin.close()
-            pretend_log_writer_process.wait()
+            if config.interactive:
+                docker_run_args = ['-it'] + docker_run_args
+
+            try:
+                announce_and_call(['docker', 'run'] + docker_run_args,
+                                  stdout=pretend_log_writer_process.stdin)
+                os.remove(pretend_fail_log_filename)
+                with suppress(OSError):
+                    os.rmdir(host_pretend_logdir_category_package)
+                    os.rmdir(host_pretend_logdir_category)
+            finally:
+                pretend_log_writer_process.stdin.close()
+                pretend_log_writer_process.wait()
 
 
 def main():
